"""
This module should use pipes to ffmpeg and async functions to easily record the screen
"""

import asyncio
import os
import subprocess as sp
import time
from abc import ABC, abstractmethod
from asyncio import Event
from contextlib import redirect_stdout

import cv2
import numpy as np
import pygame as pg
from pygame import freetype
from pygame.surface import Surface

# time helper functions; 

def _get_time(): return time.perf_counter()
_time_since = lambda t: _get_time()-t
_sleep = lambda t: time.sleep(t)

class _ScreenPiper(ABC):
    @abstractmethod
    def __init__(self, fps: float, filename: str, s: Surface = None):
        raise NotImplementedError()

    @abstractmethod
    def start(self):
        raise NotImplementedError()
    
    @abstractmethod
    def stop(self): 
        raise NotImplementedError()

class FFMPEGPiper(_ScreenPiper):
    # Vastly copied from pygame-screen-recorder
    def __init__(self, fps: float, filename: str, s: Surface = None):
        self.fps = fps
        self.filename = filename
        self.s = s or pg.display.get_surface()

        self.dtf = 1/fps # in secs
        self.pipe = None
        self.stopped = False
        self.finished_event = Event()
    
    async def start(self):
        width, height = self.s.get_size()
        size = f"{width}x{height}"
        with open(os.devnull, 'w') as nul, redirect_stdout(nul):
            self.pipe = asyncio.Popen(['ffmpeg',
                '-v','fatal',
                '-y', 
                '-f', 'rawvideo',
                '-vcodec','rawvideo',
                '-s', size,
                '-pix_fmt','rgba',
                '-r',str(self.fps),
                '-i','-',
                '-an',
                '-vcodec', 'qrtl',
                    self.filename], stdin=sp.PIPE, stdout=nul)
        asyncio.ensure_future(self.loop())

    def stop(self):
        if self.pipe is not None:
            asyncio.get_event_loop().run_until_complete(self._stop())
        else:
            raise ValueError("The Piper was stopped without being started")

    async def _stop(self):
        self.stopped = True
        await self.finished_event.wait()
        
        self.pipe.stdin.close()
        self.pipe.wait()

    async def loop(self):
        while True:
            print("looping")
            if self.stopped: 
                return self.finished_event.set()
            before = _get_time() # ns
            await self.push_frame()
            wait = max(0,self.dtf-(_time_since(before)))
            time.sleep(wait)


    async def push_frame(self):
        r = pg.surfarray.pixels_red(self.s)
        g = pg.surfarray.pixels_green(self.s)
        b = pg.surfarray.pixels_blue(self.s)
        a = np.ones_like(r)*255

        #merge the rgba channels
        mergedImage = cv2.merge((r,g,b,a))

        #rotate the image 90 degrees
        rows,cols,draws = mergedImage.shape
        M = cv2.getRotationMatrix2D((cols/2,rows/2),-90,1)
        rotatedMergedImage = cv2.warpAffine(mergedImage,M,(cols,rows))

        #flip the image horizontally
        rotatedMergedImage = cv2.flip(rotatedMergedImage,1)

        #create a mask of black pixels
        black = rotatedMergedImage[:,:,0] == 0

        #apply the mask to the frame buffer
        rotatedMergedImage[black] = [0,0,0,0]

        #feed the frame buffer to the ffmpeg pipe
        self.pipe.stdin.write(rotatedMergedImage)
        
        r = []
        g = []
        b = []
        a = []
        

async def main():
    pg.init()
    WINDOW_WIDTH, WINDOW_HEIGHT = screenSize = pg.Vector2(400,400)
    screen = pg.display.set_mode(screenSize)

    clock = pg.time.Clock()
    font = freetype.SysFont("MonoLisa", size = 30)
    counter = 0

    piper = FFMPEGPiper(60,"newvideo.mov")
    await piper.start()

    try:
        while True:
            clock.tick(10)
            if any(event.type == pg.QUIT for event in pg.event.get()):
                return
            screen.fill("black")
            font.render_to(screen,(175,175),str(counter),"green")
            counter+=1
            pg.display.flip()
            if counter>500:
                return

    finally:
        pg.quit()
        await piper._stop()

if __name__ == '__main__':
    # demo
    asyncio.run(main())



import pygame as pg
from pygame import surfarray
import numpy as np
import cv2

pg.init()

WHITE = 255, 255, 255

def show_image(image:pg.Surface|np.ndarray):
    pg.display.set_caption("Showing Image")

    if type(image)==pg.Surface:
        rect = image.get_rect()
        size = rect.size
        SCREEN = pg.display.set_mode(size,depth=32)
        SCREEN.blit(image,rect)
    elif type(image)==np.ndarray:
        size = image.shape[0:2]
        SCREEN = pg.display.set_mode(size,depth=32)
        surfarray.blit_array(SCREEN,image)

    pg.display.flip()

    running=True
    while running:
        for event in pg.event.get():
            if event.type == pg.QUIT: 
                running=False
            if event.type == pg.KEYDOWN:
                if event.key == pg.K_ESCAPE:
                    running=False

def create_nice_shape():
    #from: https://stackoverflow.com/questions/41168396/how-to-create-a-pygame-surface-from-a-numpy-array-of-float32
    x = np.arange(0, 300)
    y = np.arange(0, 300)
    X, Y = np.meshgrid(x, y)

    Z = X + Y
    #Z looks like this (if it wasn't (300,300) but (4,4)): 
    """
    [0,1,2,3]
    [1,2,3,4]
    [2,3,4,5]
    [3,4,5,6]
    """

    Z:np.ndarray = 255*Z/Z.max() #normalize Z to 255
    return Z

#from: https://gist.github.com/radames/1e7c794842755683162b
cv2_to_pg = lambda arr: cv2.cvtColor(arr.swapaxes(0,1), cv2.COLOR_BGR2RGB)

pg_to_cv2 = lambda arr: cv2.cvtColor(arr.swapaxes(0,1), cv2.COLOR_RGB2BGR)

# from numpy to pygame surface
# array = create_nice_shape()
# show_array_double(array)

# pure_cvarray=cv2.imread('crow_test.jpg',cv2.IMREAD_UNCHANGED)

# pgarray = surfarray.array3d(pg.image.load('crow_test.jpg'))

# cv2.imshow("win",pg_to_cv2(pgarray))

# array = cv2_to_pg(pure_cvarray)

# show_image(array)

# cv2.destroyAllWindows()



import time
from contextlib import contextmanager

import cv2
import numpy as np
import pygame as pg
from pygame import surfarray

pg.init()
WHITE = (255,)*3
size = (1920, 816)

@contextmanager
def timer(text):
    start = time.time()
    try:
        yield
    finally:
        end = time.time()
        print(text, end-start)

cv2_to_pg = lambda arr: cv2.cvtColor(arr.swapaxes(0,1), cv2.COLOR_BGR2RGB)

def play_video_file(filename: str = "saved_files/Narnia_01_Original.mp4"):
    video = cv2.VideoCapture(filename)
    if (video.isOpened()):
        pg.display.set_caption("Showing Video")
        SCREEN = pg.display.set_mode(size,depth=32)
        running = True
        clock = pg.time.Clock()
        with timer("Finished playing"):
            while running:
                clock.tick(60)
                for event in pg.event.get():
                    if event.type == pg.QUIT: 
                        running=False
                    if event.type == pg.KEYDOWN:
                        if event.key == pg.K_ESCAPE:
                            running=False
                running,cv_frame = video.read()
                if not running: break
                pg_frame=cv2_to_pg(cv_frame)
                surfarray.blit_array(SCREEN,pg_frame)
                pg.display.flip()
    video.release()


def gen_frames(npz: np.lib.npyio.NpzFile):
    """Supplies the numpy frames and delays the yield to maintain the correct delta time between frames. """
    def get_time(): # in ms
        return time.time()*1000
    def time_since(t): # in ms
        return get_time()-t
    files=npz.files
    dts = npz["dts"]
    assert files.pop(0) == "dts", "First array must be dts"
    last = get_time()
    for k,dt in zip(files,dts):
        pg_array = npz[k]
        diff = dt - time_since(last)
        print(k,dt,diff)
        if diff > 0:
            pg.time.delay(int(diff))
        last = get_time()
        yield pg_array

def play_frames(filename: str = "screen_recording.npz"):
    npzFile: np.lib.npyio.NpzFile = np.load(filename)
    #print(np.info(npzFile))
    pg.display.set_caption("Showing Video")
    SCREEN = pg.display.set_mode(size,depth=32)
    with timer("Finished playing"):
        for frame in gen_frames(npzFile):
            if frame is None:break
            surfarray.blit_array(SCREEN,frame)
            for event in pg.event.get():
                if event.type == pg.QUIT: break
                if event.type == pg.KEYDOWN:
                    if event.key == pg.K_ESCAPE: break
            pg.display.flip()

play_video_file()
#play_frames()

pg.quit()
#np.savez()



from collections import defaultdict
import json
import logging
import random
import time
from typing import Any, Callable, Iterable, Optional
from dataclasses import dataclass
import pygame as pg

################################################################################################
# helpers

def skip_none(x,y):
    return y if x is None else x


def dfilter(
    d: dict,
    *,
    key: Optional[Callable[[Any], bool] | list] = None,
    value: Optional[Callable[[Any], bool] | list] = None,
):
    """Filter a dictionary by key or value. You must specify either key or value"""
    if (key is None) ^ (value is None):  # xor
        if value is None:
            if isinstance(key, list):
                key = key.__contains__
            return {k: v for k, v in d.items() if key(k)}  # type: ignore
        elif key is None:
            if isinstance(value, list):
                value = value.__contains__
            return {k: v for k, v in d.items() if value(v)}
    else:
        raise TypeError("Exactly one of key and value have to be None")


start_lambda = lambda starts: lambda k: any((k.startswith(s) for s in starts))


def starts_with(s: Iterable[str] | str):
    """Returns all the pg.locals that start with any of the given strings"""
    try:
        return dfilter(pg.__dict__, key=start_lambda(s)).values()
    except TypeError:
        return dfilter(pg.__dict__, key=start_lambda([s])).values()


time_since = lambda t: time.time() - t


def false_lambda():
    return False


def dict_lambda():
    return {}


def read(path: str):
    with open(path, "r") as f:
        return json.load(f)


def write(object: Any, path: str):
    with open(path, "w") as f:
        json.dump(object, f, ensure_ascii=False)


def find(pred: Callable[[Any], bool], l: Iterable) -> Any:
    for x in l:
        if pred(x):
            return x


def print_names(l: Iterable[pg.event.Event] | Iterable[int]):
    try:
        for e in l:
            print(pg.event.event_name(e.type))  # type: ignore
    except (AttributeError, TypeError):
        for e in l:
            print(pg.event.event_name(e))  # type: ignore


def safe_item_set(d: dict):
    conv_l_t = lambda l: tuple(l) if type(l) is list else l
    return {(k, conv_l_t(v)) for k, v in d.items()}


################################################################################################


################################################################################################
# Typing
@dataclass
class SavedEvent:
    type: int
    attrs: dict
    time: float


################################################################################################


################################################################################################
# compressions
defaults: dict["str", dict["str", Any]] = {
    "window": {
        "value": None,
        "events": [
            32774,
            32785,
            770,
            32776,
            32783,
            1024,
            1025,
            1026,
            768,
            771,
            769,
            32784,
            32786,
            32777,
            32787,
        ],
    },
    "touch": {"value": False, "events": [1024]},
    "buttons": {"value": [0] * 3, "events": [1024]},  #  mouse move
    "mod": {"value": 0, "events": [771, 769]},  # Key up/key down
}


def_lookup: dict[int, dict[str, Any]] = defaultdict(dict)

for attr, d in defaults.items():
    for event in d["events"]:
        def_lookup[event][attr] = d["value"]


def single_compress(se: SavedEvent):
    if (default := def_lookup.get(se.type)) is not None:
        eit = safe_item_set(se.attrs)
        attrs = {k: v for k, v in (eit - (eit & safe_item_set(default)))}
        return se.type, attrs, se.time
    else:
        return se.type, se.attrs, se.time


def single_decompress(se: SavedEvent):
    if (d := def_lookup.get(se.type)) is not None:
        se.attrs |= d
    return se


################################################################################################


################################################################################################

old_events = {
    "mouse_foc": pg.mouse.get_focused,
    "mouse_pos": pg.mouse.get_pos,
    "mouse_rel": pg.mouse.get_rel,
    "mouse_pressed": pg.mouse.get_pressed,
    "key_foc": pg.key.get_focused,
    "key_pressed": pg.key.get_pressed,
    "key_mods": pg.key.get_mods,
}


class EventRegister:
    """
    As the name suggests this class registers the pygame events so that you can then save them.
    """

    starts = {"MOUSE", "KEY", "WINDOW"}

    def __init__(
        self,
        in_out: str = "in",
        path: Optional[str] = None,
        events: Optional[Iterable[int]] = None,
        no_catch_events: Optional[Iterable[int]] = None,
    ):
        default_events = set(starts_with(self.starts))
        self.events: set[int] = set(skip_none(events, default_events)) | {pg.QUIT}
        self.no_catch_events: set[int] = set(skip_none(no_catch_events, set())) | {
            pg.QUIT
        }
        assert in_out in ("in", "out"), 'in_out has to be either "in" or "out"'
        self._type = in_out
        self.path: Optional[str] = path
        self.random_seed = None
        self.start = time.time()
        self.rec_events: list[SavedEvent] = []
        self._reci = 0
        if self._type == "out":
            assert isinstance(self.path, str), "path must be a string"
            r = read(self.path)
            self.random_seed = r["random_seed"]
            self.rec_events = [single_decompress(SavedEvent(*x)) for x in r["events"]]
            self._focus = pg.mouse.get_focused()
            self._mouse_pos = pg.mouse.get_pos()
            self._mouse_rel = pg.mouse.get_rel()
            self._mouse_pressed = list(pg.mouse.get_pressed())
            self._key_pressed: dict = defaultdict(false_lambda)
            for i, pressed in enumerate(pg.key.get_pressed()):
                self._key_pressed[i] = pressed
            self._key_mod = pg.key.get_mods()
            pg.mouse.get_focused = self.focused
            pg.mouse.get_pos = self.mouse_pos
            pg.mouse.get_rel = self.mouse_rel
            pg.mouse.get_pressed = self.mouse_pressed  # type: ignore
            pg.key.get_focused = self.focused
            pg.key.get_pressed = self.key_pressed
            pg.key.get_mods = self.key_mods

    def _add_events(self, events: Iterable[pg.event.Event]):
        assert self._type == "in", "Can only add events when inputting"
        for e in events:
            if (t := e.type) in self.events:  # python 3.10
                self.rec_events.append(
                    SavedEvent(t, e.__dict__, time_since(self.start))
                )

    def _get_events(self) -> list[pg.event.Event]:
        assert self._type == "out", "Can only get events when outputting"
        this_events: list[pg.event.Event] = []
        for e in self.rec_events[self._reci :]:
            if time_since(self.start) >= e.time:
                self._reci += 1
                this_events.append(pg.event.Event(e.type, e.attrs))
                if e.type == pg.MOUSEMOTION:
                    self._mouse_pos = e.attrs["pos"]
                    self._mouse_rel = e.attrs["rel"]
                elif e.type == pg.MOUSEBUTTONDOWN:
                    try:
                        self._mouse_pressed[e.attrs["button"] - 1] = True
                    except IndexError:
                        pass
                elif e.type == pg.MOUSEBUTTONUP:
                    try:
                        self._mouse_pressed[e.attrs["button"] - 1] = False
                    except IndexError:
                        pass
                elif e.type in (pg.KEYDOWN, pg.KEYUP):
                    self._key_mod = e.attrs["mod"]
                    if e.type == pg.KEYDOWN:
                        self._key_pressed[e.attrs["key"] - self._key_mod] = True
                    elif e.type == pg.KEYUP:
                        self._key_pressed[e.attrs["key"] - self._key_mod] = False
                elif e.type == pg.WINDOWFOCUSGAINED:
                    self._focus = True
                elif e.type == pg.WINDOWFOCUSLOST:
                    self._focus = False
            else:
                return this_events
        return this_events

    def seed(
        self, random_seed: int | float | str | bytes | bytearray = 0
    ) -> "EventRegister":
        """
        Seed the `EventRegister` a random seed.
        Just call this on the instance to have a work around for random games (ie. most games).
        If you pass a non-faulty `random_seed` to this then it will overwrite the instances current `random_seed`.
        But in most cases just don't pass anything.
        """
        if self._type == "in":
            self.random_seed = random_seed or random.random()
        else:
            self.random_seed = random_seed or self.random_seed
        random.seed(self.random_seed)
        return self

    def get_events(self) -> list[pg.event.Event]:
        """Get the next events. To use an `EventRegister`, call this instead of `pg.event.get()`"""
        events = pg.event.get()
        if self._type == "out":
            return [
                *filter(lambda e: e.type in self.no_catch_events, events)
            ] + self._get_events()
        elif self._type == "in":
            self._add_events(events)
            return events
        else:
            raise RuntimeError('self._type must be "out" or "in"')

    def save(self):
        """Wraps up the `EventRegister`. You must call this at the end of your script."""
        if self._type == "in":
            if not self.rec_events[-1].type == pg.QUIT:
                logging.debug("Saving but PyGame hasn't finished yet")
                self.rec_events.append((pg.QUIT, {}, self.rec_events[-1][2] + 0.1))
            events = [single_compress(x) for x in self.rec_events]
            write({"events": events, "random_seed": self.random_seed}, self.path)
        else:
            pg.mouse.get_focused = old_events["mouse_foc"]
            pg.mouse.get_pos = old_events["mouse_pos"]
            pg.mouse.get_rel = old_events["mouse_rel"]
            pg.mouse.get_pressed = old_events["mouse_pressed"]
            pg.key.get_focused = old_events["key_foc"]
            pg.key.get_pressed = old_events["key_pressed"]
            pg.key.get_mods = old_events["key_mods"]

    # class getters

    def focused(self):
        return self._focus

    def mouse_pos(self):
        return self._mouse_pos

    def mouse_rel(self):
        return self._mouse_rel

    def mouse_pressed(self):
        return tuple(self._mouse_pressed)

    def key_pressed(self):
        return self._key_pressed

    def key_mods(self):
        return self._key_mod


################################################################################################


__all__ = ["EventRegister"]

################################################################################################

if __name__ == "__main__":
    print("EventRegister")



import sys,os
os.chdir(os.path.dirname(__file__))
sys.path.append("../")

from EventRegister import read, single_compress, single_decompress, json

def test_compress():
    try:
        uncomp = read("events.json")["events"]
        comp=[single_compress(ev) for ev in uncomp]
        decomp = [single_decompress(ev) for ev in comp]
        print()
        print("File:", len(json.dumps(uncomp)))
        print("Compressed:", comp_size:=len(json.dumps(comp)))
        print("Decompressed:", decomp_size:=len(json.dumps(decomp)))
        print("Compression_rate:", str(comp_rate:=(comp_size/decomp_size)*100)[:2]+"%","(Pretty Good)" if comp_rate <= 60 else "")
    except FileNotFoundError:
        print("Test compress failed: Likely file 'events.json' missing in current working directory")

if __name__ == "__main__":
    test_compress()



import os
import time
from contextlib import contextmanager

import cv2
import numpy as np
import pygame as pg
from pygame import surfarray


def pg_to_cv2(cvarray:np.ndarray)->np.ndarray:
    cvarray = cvarray.swapaxes(0,1) #rotate
    cvarray = cv2.cvtColor(cvarray, cv2.COLOR_RGB2BGR) #RGB to BGR
    return cvarray

def timer_wrapper(func):
    def inner(*args, **kwargs):
        start = time.time()
        func(*args, **kwargs)
        end = time.time()
        #print("Finished:" ,func.__name__ ,end-start)
        return end - start
    return inner

@contextmanager
def timer(text):
    start = time.time()
    try:
        yield
    finally:
        end = time.time()
        print(text, end-start)


@contextmanager
def video_writer(*args,**kwargs):
    video = cv2.VideoWriter(*args,**kwargs)
    try:
        yield video
    finally:
        video.release()

def surf_to_arr(frame: pg.Surface):
    try:
        pg_frame = surfarray.pixels3d(frame) # convert the surface to a np array. Only works with depth 24 or 32, not less
    except:
        pg_frame = surfarray.array3d(frame) # convert the surface to a np array. Works with any depth
    return pg_frame


@timer_wrapper
def save_video(frames: list, average_dt: float|list, file_type: str = "mp4", name: str = "screen_recording"):
    if type(average_dt) is list: average_dt = sum(average_dt)/len(average_dt) # force average_dt to be a float
    size = frames[0].get_size()
    codec_dict={
        "avi":'DIVX',
        "mp4":'MP4V'
    }
    codec = cv2.VideoWriter_fourcc(*codec_dict[file_type])
    with video_writer(name+"."+file_type, codec, 1000/average_dt, size) as video: # file_name, codec, average_fps, dimensions
        for frame in frames:
            pg_frame = surf_to_arr(frame)
            cv_frame = pg_to_cv2(pg_frame)  # then convert the np array so it is compatible with opencv
            video.write(cv_frame)   #write the frame to the video using opencv

@timer_wrapper
def save_frames(frames: list, dts: list, name: str = "screen_recording"):
    # instead of saving the images as a video, like any normal programmer, we save them as numpy images. 
    pg_arrays = (surf_to_arr(frame) for frame in frames)
    with timer("Finished saving"):
        np.savez_compressed(name+".npz",*pg_arrays,dts=np.array(dts),)

def draw_fps(s:pg.Surface,clock:pg.time.Clock): 
    fps = clock.get_fps()
    sysfont.render_to(s,(100,100),str(fps),fgcolor=nice_green)

# initializing globals (colors, fonts, window, etc.)
pg.init()
sysfont = pg.freetype.SysFont(None,40)
BLACK = (0,)*3
nice_green = pg.Color("chartreuse2")
size=(1000, 600)
pg.display.set_caption("Screen Recording")
window = pg.display.set_mode(size)
# this is to save the frames
frames = []
dts = []
clock = pg.time.Clock()

running=True
try:
    while running:
        dt = clock.tick(60) # aim for ... fps
        for event in pg.event.get():
            if event.type == pg.QUIT:
                running=False
            if event.type == pg.KEYDOWN:
                if event.key == pg.K_ESCAPE:
                    running=False
        window.fill(BLACK)
        draw_fps(window,clock)
        window_copy = window.copy() # if we don't copy the window then we will have the same image in all frames at the end
        frames.append(window_copy) # We save the current frame together with the time passed between the frames
        dts.append(dt)
        pg.display.flip()
        if len(frames) >= 100: running = False # uncomment this to stop the game after ... frames for similar results in every run"
finally:
    pg.quit()
    # At this stage, the game ended and the cleanup process can start. For this we convert the frames to opencv images
    # Only then we will write the video to the hard drive (That is what makes the save so slow).
    # General information about the recording
    frame_num = len(frames)
    dt_sum = sum(dts)
    average_dt = dt_sum/frame_num
    # This is only an approximation: 
    # for each frame we have width * height many pixels -> frame_num * width * height
    # A Surface needs get_bytesize() many bytes per pixel (In this case 4 bytes, because we set the depth of the display to 32 bits)
    memory_usage_approx = frame_num * size[0] * size[1] * frames[0].get_bytesize()  #https://www.pygame.org/docs/ref/surface.html#pygame.Surface.get_bytesize
    print("Total time:" , dt_sum/1000,"s")
    print("Average time per frame:" , average_dt,"ms")
    print("Memory usage approximation" , memory_usage_approx/1000, "KB")
    # args = (frames,dts,"avi","screen_recording")
    # time_for_save = save_video(*args)
    # file_name = args[3]+"."+args[2]
    args = (frames, dts, "screen_recording")
    time_for_save = save_frames(*args)
    file_name = args[2]+".npz"
    disk_memory_usage = os.path.getsize(file_name)
    print("Video memory usage:" , disk_memory_usage/1000, "KB")
    with open("test.txt", "a") as f:
        print("Total video length:" , dt_sum/1000,"s\nNumber of frames:", frame_num,"\nFPS:",1000/average_dt,"\nSize:",size,"\nTime for save:",time_for_save,"s\nSaved in file:",file_name,"with",disk_memory_usage/1000, "KB memory",file=f)
        print("_"*100,file=f)



import atexit, itertools, logging, os, time, threading
import math
from contextlib import contextmanager
from functools import cache, wraps
from threading import Thread
from types import GenericAlias
from typing import (Callable, Coroutine, List, Literal, Optional,
                    SupportsIndex, Tuple, Type, _GenericAlias,
                    _UnionGenericAlias)

import cv2
import numpy as np
import pygame as pg

from ScreenRecorder import AnyPath, Anypath_type, available_formats, PlayingThread, Preloader


class Preloader(PlayingThread): 
    def __init__(self, get_frames: Callable[[int],pg.Surface],thread: PlayingThread, max_preload:int = 100):
        self.get_frames = get_frames
        self.thread = thread
        self.max_preload = max_preload
    def run(self): 
        while self.running:
            if self.position-self.thread.position<self.max_preload:
                try:
                    self.get_frames(self.position)
                    self.position+=1
                except (IndexError):
                        break



_props = {
    "width":cv2.CAP_PROP_FRAME_WIDTH,
    "height":cv2.CAP_PROP_FRAME_HEIGHT,
    "fps":cv2.CAP_PROP_FPS,
    "frame_count":cv2.CAP_PROP_FRAME_COUNT
}

video_properties = list(_props.keys())+["size","all"]
""" The list of properties a video may have """

class Video:
    """ Represents a Video Source. """
    def __init__(self,source: AnyPath | int):
        """ Inits a Video
        
        Parameters
        ----------
        source: AnyPath | int | cv2.VideoCapture 
            If path, opens that file as a stream 
            if int, then opens the i-th camera as stream.
        """
        self.cap = cv2.VideoCapture(source)
        if not self.cap.isOpened():
            try:
                with open(source) as _: 
                    if type(source) is str: 
                        assert source.split(".")[-1] in available_formats # Test if file can be opened
            except:
                problem = "File cannot be opened currently" if os.path.isfile(source) else "File/Camera doesn't exist/cannot be opened"
                raise ValueError(problem+": "+str(source))
        assert isinstance(source,Anypath_type),("source has to be a path or an int")

    def __del__(self):
        self.cap.release()

    def play(self,*args,**kwargs)->'VideoPlayer':
        """ 
        Creates a VideoPlayer object, starts playing it, and returns it. 
        All args and kwargs will be passed to VideoPlayer
        """
        vip = VideoPlayer(self,*args,*kwargs)
        vip.play()
        return vip
    
    @cache
    def __getattr__(self, prop: Literal["width","height","fps","frame_count","size","all"] | int | str) -> int|float:
        """ Returns the given property of the video """
        if type(prop) is int:
            return self.cap.get(prop)
        elif type(prop) is str:
            cv_prop = _props.get(prop)
            if cv_prop is not None:
                result = self.cap.get(cv_prop)
                if prop == "fps":
                    return result
                else:
                    return int(result)
            elif prop == "size":
                return self.width,self.height
            elif prop == "all":
                return tuple(self.__getattr__(prop) for prop in video_properties)
            return self.__getattribute__(prop)

class VideoPlayer:
    """ 
    This is a very naive VideoPlayer. Please do not use it for large Videos (like Movies). 
    It is rather intended for small animations or little Recordings that might be replayed a lot of times
    """
    def __init__(self,source: AnyPath | Video, surf: Optional[pg.Surface] = None, on_stop: Optional[Callable[[],None]] = None, always_load: bool = True):
        """always_load(bool): Whether to always load the full video until the given position when seeking"""
        if isinstance(source,Video):
            self.video = source
        else:
            self.video = Video(source)
        self.surf = surf or pg.display.get_surface()
        self.on_stop = on_stop or mt
        self.always_load = always_load
        if not self.cap.isOpened(): raise RuntimeError("Video isn't opened")
        self.will_scale = not self.size == self.surf.get_size()
        if self.will_scale: 
            logging.warning("Video size doesn't match surface size. Will be rescaled")
            self.scale = self.surf.get_size()
        self.frames = [None]*self.frame_count
        self.max_preload = int(0.2 * math.log(len(self.frames)))
        logging.debug("Max Preload of video: "+str(self.max_preload))
        self._init_thread()
    
    def _on_stop(self):
        ret_val = self.on_stop()
        if ret_val is False: 
            self._init_thread()
        elif ret_val is True:
            self.play()

    def _init_thread(self): 
        """ Internal function to init a new thread """
        new_thread = PlayingThread(self.surf,1000/self.fps,self._get_frames,self._on_stop)
        logging.debug(f"Initing new Thread ({new_thread.name}) of {self.__class__.__name__}")
        try:
            new_thread.position = self.thread.position
        except NameError:
            pass
        self.thread= new_thread
    def _init_thread(self):        
        self.thread = PlayingThread(self.surf, 1000/self.fps, self._get_frames,self._on_stop)
        self.loader = Preloader(self._get_frames,self.thread,self.max_preload)
        
    def __getattr__(self, key):
        try:
            return self.video.__getattr__(key)
        except NameError: 
            return self.__getattribute__(key)

    @property
    def position(self): 
        return int(self.cap.get(cv2.CAP_PROP_POS_FRAMES))

    @position.setter
    def _(self)->None:
        self.cap.set(cv2.CAP_PROP_POS_FRAMES)

    def _get_frames(self, position: int):
        """ Internal function to get the frame at a specific position """
        frame = self.frames[position]
        while frame is None:
            self._get_next()
            frame = self.frames[position]
        return frame
    
    def _get_next(self)->pg.Surface:
        """ Internal function to get the next frame """
        position = self.position
        logging.debug("Getting next frame at position: "+str(position))
        ret,cv_frame = self.cap.read()
        if not ret: raise IndexError
        surf = pg.surfarray.make_surface(_cv2_to_pg(cv_frame))
        if self.will_scale:
            surf = pg.transform.scale(surf,self.scale)
        self.frames[position] = surf
        return surf

    def stop(self):
        super().stop() # stops the thread
        self.loader.stop()
        self.cap.release()

    def seek(self,position:int)->None:
        if self.always_load: self.position = position
        self.loader.position = position
        return super().seek(position)

    def reuse(self,new_on_stop: Optional[Callable[[],None]] = None)->'RecordingPlayer':
        if not self.is_("loaded"): 
            if self.is_("stopped"):
                raise RuntimeError("Reusing a not fully loaded stopped VideoPlayer.")
            else: 
                logging.warning("Reusing a not fully loaded Player. Might take a while")
                unloadPlayer(self)
                assert self.is_("loaded")
        rec = Recording(self["fps"],self["size"])
        rec.frames = self.frames
        on_stop = new_on_stop or self.on_stop
        return RecordingPlayer(rec,on_stop,self.surf)
    




import atexit, logging, os, time, threading
from string import printable
from contextlib import contextmanager
from functools import wraps
from threading import Thread
from typing import (Callable, Coroutine, List, Optional,
                    Sequence, Tuple)

import cv2
import numpy as np
import pygame as pg

##############################################################################################################
#Helpers

def mt(*_,**__): pass

#types

def ensure_type(var,_type):
    if not isinstance(var,_type):
        raise TypeError(f"{var=} must by of type {_type=}")

AnyPath = str | bytes | os.PathLike[str] | os.PathLike[bytes]

#Test utils

def timer_wrapper(func,info: str = ""):
    @wraps(func)
    def inner(*args, **kwargs):
        start = time.time()
        ret_val = func(*args, **kwargs)
        end = time.time()
        logging.debug("Finished:" ,func.__name__ ,end-start, info)
        return ret_val
    return inner


# time helper functions; all in milliseconds

def _get_time(): return time.time()*1000
_time_since = lambda t: _get_time()-t
_sleep = lambda t: time.sleep(t/1000)

# Video Converting:

_pg_to_cv2 = lambda arr: cv2.cvtColor(arr.swapaxes(0,1), cv2.COLOR_RGB2BGR)

def _surf_to_arr(frame: pg.surface.Surface):
    """ Internal function to convert a pg.surface.Surface to a np.ndarray depending on the depth. 
    See the pygame documentation for information on the surfarray submodule"""
    try:
        pg_frame = pg.surfarray.pixels3d(frame) # convert the surface to a np array. Only works with depth 24 or 32, not less
    except:
        pg_frame = pg.surfarray.array3d(frame) # convert the surface to a np array. Works with any depth
    return pg_frame


# _cleanup handler:

def cleanup():
    for k,v in globals().items():
        if isinstance(v, RecordingPlayer):
            if v.is_("playing"):
                logging.warning(f"Stopping not ended player ({k}) ...")
                v.stop()
        elif isinstance(v,ScreenRecorder):
            if v.running_thread is not None:
                logging.warning(f"Stopping not ended recording ({k}) ...")
                v.running_thread.stop()
    for thread in threading.enumerate():
        if not thread.name == "MainThread":
            logging.error("Joining unfinished Thread: " + str((thread.name,thread.__class__.__name__)))
            thread.join()


##############################################################################################################


##############################################################################################################
#Compression

def compress(frame: pg.surface.Surface, comp_rate: int)->pg.surface.Surface:
    if comp_rate:
        vec = pg.Vector2(frame.get_size())
        return pg.transform.smoothscale(frame,vec/(2**comp_rate))
    else:
        return frame

def decompress(frame: pg.surface.Surface,compress: int):
    for _ in range(compress): 
        frame = pg.transform.scale2x(frame)
    return frame

##############################################################################################################


##############################################################################################################
# Threading

class StoppableThread(Thread):
    """ A general StoppableThread. It can be stopped with thread.stop()"""
    def __init__(self,coro_make: Callable[[],Coroutine]):
        """ The coro_make is a Function that returns a Coroutine """
        Thread.__init__(self)
        self.coro = coro_make()
        self.running: bool = True

    def run(self):
        logging.debug("Started "+self.__class__.__name__)
        while self.running:
            self.coro.__next__()
        self.coro.close()
        logging.debug("Finished "+self.__class__.__name__)
    
    def stop(self):
        self.running = False

class RecordingThread(StoppableThread):
    def __init__(self,surf: pg.surface.Surface, dt: float, save_frames: Callable[[pg.surface.Surface],None]):
        def coro():
            last = _get_time()
            try:
                while True:
                    yield
                    save_frames(surf.copy())
                    diff = dt - _time_since(last)
                    if diff > 0:
                        _sleep(diff)
                    last = _get_time()
            except (pg.error, GeneratorExit):
                pass
        super().__init__(coro) # give the coroutine to the StoppableThread init

class PlayingThread(StoppableThread):
    """ Intern Thread for playing Recordings """
    def __init__(self,surf: pg.surface.Surface, dt: float, get_frames: Callable[[int],pg.surface.Surface], on_stop: Callable[[],None]):
        """ The init for a PlayingThread
        
        Parameters
        ----------
        surf: pg.surface.Surface
            The Surface to draw on
        dt: float
            The milliseconds between each frame
        get_frames: (int)->pg.surface.Surface
            The function to get the frame at a specified position
        on_stop: ()->None
            The callback to call when the Thread stops playing
        """
        Thread.__init__(self)
        self.running = True
        self.paused = False
        self.position = 0
        self.surf = surf
        self.dt = dt
        self.get_frames = get_frames
        self.on_stop = on_stop

    def run(self):
        last = _get_time()
        while self.running:
            if not self.paused:
                try:
                    frame = self.get_frames(self.position)
                    self.surf.blit(frame, (0,0))
                except (IndexError, pg.error):
                    break
                self.position+=1
                diff = self.dt - _time_since(last)
                if not self.running: break
                if diff > 0:
                    _sleep(diff)
                last = _get_time()
        self.running = False
        self.on_stop()
        
    def play(self):
        self.paused = False

    def pause(self):
        self.paused = True

##############################################################################################################


##############################################################################################################
#Recording

class Recording:
    save: Callable
    def __init__(self, fps: float, size: Tuple[int,int], compress: int):
        self.fps = fps
        self.size  = size
        self.compress = compress
        self.dt = 1000/fps
        self.frames: List[pg.surface.Surface] = []
        self.__getitem__ = self.frames.__getitem__ # type: ignore
    
    def add_frame(self,frame: pg.surface.Surface):
        """ Add a frame to the recording"""
        self.frames.append(compress(frame,self.compress))

    @property
    def frame_number(self):
        """ Number of frames in the recording"""
        return len(self.frames)

    @property
    def total_length(self):
        """ Length of the recording in seconds """
        return len(self.frames)*self.dt/1000

    def resize(self,size: Tuple[int,int]):
        """ Resize a Recording """
        if not size == self.size:
            self.frames = [pg.transform.smoothscale(frame,size) for frame in self.frames]
            self.size = size
        return self

    def apply(self,effect: Callable[[pg.surface.Surface], pg.surface.Surface]):
        self.frames = [effect(frame) for frame in self.frames]
        return self

    def __getitem__(self,index):
        try:
            self.frames = [*self.frames[index]] # slice
        except TypeError:
            #self.frames = [self.frames[index]] # index
            raise TypeError("You can only slice frames")
        return self

    def __sizeof__(self) -> int:
        try:
            byte_size = self.frames[0].get_bytesize()
            w,h = self.size
            comp_fact = (self.compress**2) or 1
            return int(self.frame_number * w * h * byte_size / comp_fact)
        except IndexError: 
            return 0
 

class ScreenRecorder:
    def __init__(self, fps: float = None, surf: Optional[pg.surface.Surface] = None, compress: int = 0, stream = None):
        """ Inits a ScreenRecorder
        @params:
        fps: float 
            Frames per second the screen recorder should record (default=60)
        surf: pg.surface.Surface 
            The surface that should be recorded (default=pg.display.get_surface())
        compress: int = 0
            Whether to compress the recording and by how much. 
        stream: Implements .send() 
            A stream that gets a Recording object when the recording is stopped with recorder.stop_rec_to_stream()
        """
        self.fps: float = fps or 60
        self.surf = surf or pg.display.get_surface()
        self.compress = compress
        self.stream = stream
        assert self.surf is not None, "You haven't set the display mode yet or you haven't inited pygame!"
        self.size = self.surf.get_size()
        self.dt: float = 1000/self.fps
        self.running_recording: Optional[Recording] = None
        self.running_thread: Optional[RecordingThread] = None
        self.recordings: List[Recording] = []

    def start_rec(self,fps: float = None):
        """ Starts a recording
        @params:
        fps: overrides the classes default fps for this recording
        """
        assert self.running_thread is None, "You can't start recording without stopping the current recording."
        self.running_recording = Recording(fps or self.fps,self.size,self.compress)
        self.running_thread = RecordingThread(self.surf, self.running_recording.dt, self.running_recording.add_frame)
        self.running_thread.start()
        return self
    
    def _stop_rec(self):
        if self.running_thread is not None:
            self.running_thread.stop()
            rec = self.running_recording
            self.running_recording=None
            self.running_thread = None
            return rec
    
    def stop_rec(self):
        """ Stops the current recording """
        rec = self._stop_rec()
        self.recordings.append(rec)
        if self.stream is not None:
            self.stream.send(rec)
        return self

    def stop_rec_to_stream(self,stream = None):
        """ Stops the current recording and calls the given stream's send method with the recording 
        @params:
        stream: Overrides the classes default stream for this function call
        """
        stream = stream or self.stream
        assert stream is not None, "You haven't set a stream yet"
        rec = self._stop_rec()
        stream.send(rec)
        return self

    def abort_rec(self):
        """ Aborts the recording (Stops the recording without saving it) """
        self._stop_rec()
        return self

    def get_single_recording(self):
        """ 
        This gets the first recording in the list of recordings. 
        In many cases you just have one recording and then you get that one 
        """
        assert len(self.recordings)>0, "You haven't finished a recording yet"
        return self.recordings[0]

    def get_recordings(self):
        """
        Get all recordings until now
        """
        return self.recordings

    @timer_wrapper
    def save_recordings(self, key: str | Sequence[tuple[str,str]] | Callable[[int],tuple[str,str]], save_dir: AnyPath = None, blocking: bool = True):
        """ Inits a RecordingSaver with the recordings and calls its save function with the given arguments.  
        @params:
        key (str | Sequence[tuple[str,str]] | Callable[[int],tuple[str,str]]): If the key is a str then it is the format in which the recordings are saved.
        Otherwise the key will be given the index of the recording and should return a valid (filename,extension) tuple for each recording. 
        If the key instead returns None the recording will be skipped. 
        save_dir (AnyPath): The directory where to save the recordings. Takes anything that os.chdir takes. Defaults to ""
        blocking (bool): Whether to block the calling process until the recordings are all saved or not. Easiest and default is True
        @returns
        If the RecordingSaver was set to blocking then this returns a list of filepaths. 
        Otherwise it returns a function that returns that list and **has** to be called before the program ends.
        """
        if not self.recordings: 
            logging.debug("No recordings to save")
        saver = RecordingSaver(self.recordings, key, save_dir, blocking)
        return saver.save()

##############################################################################################################


##############################################################################################################
# Saving

codec_dict={
    "avi":'DIVX',
    "webm":'WEBM',
    "mp4":'MP4V'
}

try:
    _codec_dict = {key:cv2.VideoWriter_fourcc(*v) for key,v in codec_dict.items()}
except:
    logging.warning("FFMPEG is not available. Install it to save your recording as a video")

def add_codec(format:str, codec: int | str):
    if type(codec) is str: 
        codec = cv2.VideoWriter_fourcc(*codec)
    _codec_dict[format] = codec


available_formats = [*codec_dict.keys(),"npz"]
""" The available formats to save a recording"""
logging.debug("Available formats: "+", ".join(available_formats))

@contextmanager
def _video_writer(path: str, codec: int, fps: float, dimensions: Tuple[int, int]):
    """ Internal contextmanager that automatically handles the opening and closing of a cv2.VideoWriter """
    video = cv2.VideoWriter(path, codec, fps, dimensions)
    try:
        yield video
    finally:
        video.release()

def save_recording(rec: Recording,ftuple:Tuple[str,str])->str:
    """ Saves a single recording given a recording and a (filename,extension) tuple. Blocks until the recording is saved. 
    
    Parameters
    ----------
    rec: Recording 
        The recording to save
    ftuple: Tuple[str,str]
        A (filename, extension) tuple. Available extensions are in the documentation
    
    Returns
    -------
    The path of the saved recording

    Raises
    ------
    TypeError
        If rec is not of type Recording
    RuntimeError
        If a video is being saved and ffmpeg is not installed
    """
    ensure_type(rec,Recording)
    logging.debug("Saving recording to: "+os.path.abspath(".".join(ftuple)))
    frames: List[np.ndarray] = [_surf_to_arr(decompress(frame,rec.compress)) for frame in rec.frames]
    _save_single(frames,rec.fps,rec.size,*ftuple)
    return ".".join(ftuple)

Recording.save = save_recording

def _save_single(frames: List[np.ndarray], fps: float, size: Tuple[int,int], fname: str, ext: str):
    path = f"{fname}.{ext}"
    if ext == "npz":
        return _save_as_np(frames, fps, size, path)
    elif codec_dict.get(ext):
        return _save_as_video(frames, fps, size, path, ext)
    raise ValueError ("Extension not supported: "+ext)

def _save_as_video(frames: List[np.ndarray], fps: float, size: Tuple[int,int], path: str, ext: str):
    try:
        codec = _codec_dict[ext]
        with _video_writer(path, codec, fps, size) as video:
            for frame in frames:
                cv_frame = _pg_to_cv2(frame)  # convert the np array so it is compatible with opencv
                video.write(cv_frame)   #write the frame to the video using opencv
    except NameError as e:
        if e.name == "_codec_dict":
            raise RuntimeError("FFMPEG isn't available.")
        else: raise

def _save_as_np(frames: List[np.ndarray], fps: float, size: Tuple[int,int], path: str):
    meta = np.array([fps,*size],dtype=np.float64)
    np.savez_compressed(path,*frames,meta=meta)

class RecordingSaver:
    """ A class that is responsible for saving the recordings """
    def __init__(self, recordings: List[Recording], key: str | Sequence[tuple[str,str]] | Callable[[int],tuple[str,str]], save_dir: AnyPath = None, blocking: bool = True):
        """ Inits a RecordingSaver
        
        Parameters
        ----------
        recordings: List[Recording]
            The list of recordings to save.
        key: str | Sequence[tuple[str,str]] | Callable[[int],tuple[str,str]]
            If the key is a str then it is the format in which the recordings are saved.
            Otherwise the key will be given the index of the recording and should return a valid (filename,extension) tuple for each recording. 
            If the key instead returns None the recording will be skipped. 
        save_dir: AnyPath
            The directory where to save the recordings. Takes anything that os.chdir takes. If not given or set to None then saves in the current directory ""
        blocking: bool 
            Whether to block the calling process until the recordings are all saved or not. Easiest and default is True
        """
        def str_lambda(x: int):
            return ("recording_"+str(x),key)
        _Key = Callable[[int],Optional[Tuple[str,str]]]
        _key: _Key
        try:
            if isinstance(key, str): 
                _key = str_lambda
            elif callable(key): 
                _key = key
            else: 
                _key = key.__getitem__
        except NameError as e: 
            if e.name == "__getitem__":
                raise TypeError ("key has to be one of str, callable or SupportsIndex")
            else: 
                raise
        self.recordings = recordings
        self.key = _key
        self.save_dir = save_dir or "."
        self.blocking = blocking
        self.saved_recordings: List[str] = []
        assert os.path.isdir(self.save_dir), \
            (f'save_dir has to be a directory: ' + str(save_dir))

    def _save(self):
        """ Thread target function for saving """
        assert self.recordings is not None, "recordings have to be set before saving"
        if self.save_dir is not None:
            old_wd = os.getcwd()
            os.chdir(self.save_dir)
        threads = []
        for i,rec in enumerate(self.recordings):
            ftuple: Optional[Tuple[str,str]] = self.key(i)
            if ftuple is not None:
                thread = Thread(target=rec.save_recording,args=(ftuple))
                thread.start()
                threads.append((thread,".".join(ftuple)))
        for thread,saved_rec in threads:
            thread.join()
            self.saved_recordings.append(saved_rec)
        if self.save_dir is not None:
            os.chdir(old_wd)
        

    def save(self)->Callable[[],List[str]]|List[str]:
        """ Saves the given recordings.

        If the RecordingSaver was set to blocking then this returns a list of paths. 
        Otherwise it returns a function that returns that list and **has** to be called before the program ends.
        """
        try:
            save_thread = Thread(target=self._save,daemon=(not self.blocking))
        except:
            raise RuntimeError("Save thread failed")
        save_thread.start()
        if self.blocking:
            save_thread.join()
            return self.saved_recordings
        else:
            def func():
                save_thread.join()
                return self.saved_recordings
            return func

##############################################################################################################


##############################################################################################################
# Playing
def play_indefinite(player: 'RecordingPlayer'):
    player.restart()

class play_n_times:
    def __init__(self, n: int):
        self.n = n
    def __call__(self, player: 'RecordingPlayer'):
        self.n-=1
        if self.n>=1:
            player.restart()

class play_n_wrapper:
    def __init__(self, n: int):
        self.n = n
        self.func = None
    def __call__(self, func):
        @wraps(func)
        def on_stop(player: 'RecordingPlayer'):
            try:
                res = func(player)
            except TypeError:
                res = func()
            self.n-=1
            if res and self.n>=1:
                player.restart()
        return on_stop

def npz_player(path: AnyPath,*args,**kwargs):
    """ In most cases use `NPZPlayer`"""
    npzFile = np.load(path)
    fps,w,h = npzFile["meta"]
    size = int(w),int(h)
    rec = Recording(fps, size, compress=0)
    for i in range(len(npzFile.files)): 
        rec.add_frame(pg.surfarray.make_surface(npzFile[f"arr_{i}"]))
    return RecordingPlayer(rec,*args,**kwargs)

class RecordingPlayer:
    """ 
    A RecordingPlayer that takes your pure recordings and plays them very easily. 
    """
    def __init__(self,rec: Recording, on_stop: Optional[Callable[[],None]] = None, surf: Optional[pg.surface.Surface] = None):
        self.rec = rec
        self.on_stop = on_stop if on_stop is not None else mt
        self.surf = surf or pg.display.get_surface()
        self.rec.resize(self.surf.get_size())
        self._init_thread()

    def _on_stop(self):
        try:
            self.on_stop(self)
        except TypeError:
            self.on_stop()

    def _init_thread(self):
        """ Inits a new thread for the player"""
        if hasattr(self, 'thread'):
            self.thread.stop()
        self.thread = PlayingThread(self.surf, self.rec.dt, self._get_frame, self._on_stop)

    def _get_frame(self, pos: int): 
        frame = self.rec.frames[pos]
        frame = decompress(frame,self.rec.compress)
        return frame

    def stop(self):
        """ Forcefully stop a player even before he ends naturally. Can be called several times. """
        self.on_stop = mt
        self.thread.stop()
        self.thread.join()
        return self

    def play(self):
        """ Plays the player. Can be called multiple times. If the player was paused previously, it will resume."""
        if self.is_("stopped"):
            raise RuntimeError("""Restart a finished player with restart""")
        elif self.is_("paused"):
            self.thread.play()
        elif not self.is_("started"):
            self.thread.start()
        return self
    def pause(self):
        """ Pauses the player. Can be called multiple times. 
        Raises a RuntimeError if the player hasn't started yet or already finished"""
        if self.thread.is_alive():
            self.thread.pause()
            return self
        else:
            if self.thread.running:
                raise RuntimeError("You cannot pause before starting to play")
            else:
                raise RuntimeError("The player already finished. You cannot pause anymore. Restart a finished player with restart")
    def restart(self):
        """ Starts playing from the beginning, whether the player stopped or not """
        print("Restarting")
        if self.is_("stopped"):
            self._init_thread()
            self.thread.start()
        else:
            self.thread.position = 0
            self.thread.play()
        return self
    def seek(self, position: int):
        """ 
        Sets the current position of the player. Can be called before the player started. 
        This won't catch if the position is too high. 
        Instead the player will just finish on the next iteration. 
        Raises an TypeError if the position is not an int. 
        """
        try:
            self.thread.position = int(position)
            return self
        except TypeError:
            raise TypeError(f"Position (type:{position.__class__.__name__} couldn't be converted to an integer")
    def tell(self)->int:
        """ Returns the current position of the player """
        return self.thread.position
    def seekms(self, time: float):
        """ Goes to the specified time in ms (floor) """
        return self.seek(int(time/self.rec.dt))
    def tellms(self)->float:
        """ Convinience function: returns the current play time. `self.rec.dt * self.tell()`"""
        return self.rec.dt * self.tell()
    def is_(self,attr: str)->bool:
        """ Test the state of the player

        Parameters
        ----------
        attr (str): select what state you want to get. 
            - "started" -> whether the player is started (and still running)
            - "stopped" -> whether the player has stopped/finished
            - "playing" -> whether the player is playing
            - "paused" -> whether the player is paused
        
        Raises
        ------
        TypeError:
            If attr is not of type str
        ValueError:
            If attr is none of the above
        """
        ensure_type(attr,str)
        if attr == "started":
            return self.thread.is_alive() and self.thread.running
        elif attr == "stopped":
            return not self.thread.running
        elif attr == "playing":
            return self.thread.is_alive() and not self.thread.paused
        elif attr == "paused":
            return self.thread.is_alive() and self.thread.paused
        else:
            raise ValueError(f'attr must be either {" or ".join(["started","stopped","playing","paused"])}')

class NPZPlayer(RecordingPlayer):
    def __init__(self, path: AnyPath, on_stop: Optional[Callable[[],None]] = None, surf: Optional[pg.surface.Surface] = None):
        """ Inits a NPZPlayer 
        
        Parameters
        ----------
        path: AnyPath
            The file to be opened
        on_stop: Optional[Callable[[],None]]
            What to do when the Player finished
        surf: Optional[pg.surface.Surface]
            The Surface to play on
        """
        self.npzFile = np.load(path)
        self.surf = surf or pg.display.get_surface()
        self.on_stop = on_stop or mt
        fps,w,h = self.npzFile["meta"]
        self.fps = fps
        self.size = int(w),int(h)
        self._init_thread()

    def _init_thread(self):
        self.thread = PlayingThread(self.surf,1000/self.fps,self._get_frames,self._on_stop)
    
    def _get_frames(self,position:int)->pg.surface.Surface:
        try:
            np_arr = self.npzFile[f"arr_{position}"]
            frame = pg.surfarray.make_surface(np_arr)
            return frame
        except KeyError:
            raise IndexError


class PlayerList:
    """ Still in development: Plays a List of players """
    


##############################################################################################################


__all__ = [
    "ScreenRecorder",
    "available_formats",
    "RecordingSaver",
    "save_recording",
    "npz_player",
    "RecordingPlayer"
]

atexit.register(cleanup)

##############################################################################################################
#Testing/Demo

if __name__ == "__main__":
    print("ScreenRecorder")
        

##############################################################################################################
